@@ -2,13 +2,12 @@
 """DB-SIG compliant module for communicating with MS SQL servers"""
 
 __author__ = 'Mikhail Denisenko <denisenkom@gmail.com>'
-__version__ = '0.5.0'
+__version__ = '0.6.0'
 
 import logging
 import decimal
 import re
 from tds import *
-from tds import _TdsContext
 from login import *
 from query import *
 
@@ -30,58 +29,13 @@ DB_RES_NEXT_RESULT     = 3
 DB_RES_NO_MORE_RESULTS = 4
 DB_RES_SUCCEED         = 5
 
-def prdbresults_state(retcode):
-    if retcode == DB_RES_INIT:                 return "DB_RES_INIT"
-    elif retcode == DB_RES_RESULTSET_EMPTY:    return "DB_RES_RESULTSET_EMPTY"
-    elif retcode == DB_RES_RESULTSET_ROWS:     return "DB_RES_RESULTSET_ROWS"
-    elif retcode == DB_RES_NEXT_RESULT:        return "DB_RES_NEXT_RESULT"
-    elif retcode == DB_RES_NO_MORE_RESULTS:    return "DB_RES_NO_MORE_RESULTS"
-    elif retcode == DB_RES_SUCCEED:            return "DB_RES_SUCCEED"
-    else: return "oops: %d ??" % retcode
-
-def prretcode(retcode):
-    if retcode == TDS_SUCCESS or retcode is None:return "TDS_SUCCESS"
-    elif retcode == TDS_FAIL:                   return "TDS_FAIL"
-    elif retcode == TDS_NO_MORE_RESULTS:        return "TDS_NO_MORE_RESULTS"
-    elif retcode == TDS_CANCELLED:              return "TDS_CANCELLED"
-    else: return "oops: %u ??" % retcode
-
-
-def prresult_type(result_type):
-    if result_type == TDS_ROW_RESULT:          return "TDS_ROW_RESULT"
-    elif result_type == TDS_PARAM_RESULT:      return "TDS_PARAM_RESULT"
-    elif result_type == TDS_STATUS_RESULT:     return "TDS_STATUS_RESULT"
-    elif result_type == TDS_MSG_RESULT:        return "TDS_MSG_RESULT"
-    elif result_type == TDS_COMPUTE_RESULT:    return "TDS_COMPUTE_RESULT"
-    elif result_type == TDS_CMD_DONE:          return "TDS_CMD_DONE"
-    elif result_type == TDS_CMD_SUCCEED:       return "TDS_CMD_SUCCEED"
-    elif result_type == TDS_CMD_FAIL:          return "TDS_CMD_FAIL"
-    elif result_type == TDS_ROWFMT_RESULT:     return "TDS_ROWFMT_RESULT"
-    elif result_type == TDS_COMPUTEFMT_RESULT: return "TDS_COMPUTEFMT_RESULT"
-    elif result_type == TDS_DESCRIBE_RESULT:   return "TDS_DESCRIBE_RESULT"
-    elif result_type == TDS_DONE_RESULT:       return "TDS_DONE_RESULT"
-    elif result_type == TDS_DONEPROC_RESULT:   return "TDS_DONEPROC_RESULT"
-    elif result_type == TDS_DONEINPROC_RESULT: return "TDS_DONEINPROC_RESULT"
-    elif result_type == TDS_OTHERS_RESULT:     return "TDS_OTHERS_RESULT"
-    else: "oops: %u ??" % result_type
-
-
-class MemoryChunkedHandler(object):
-    def begin(self, column, size):
-        logger.debug('MemoryChunkedHandler.begin(sz=%d)', size)
-        self.size = size
-        self.sio = StringIO()
-    def new_chunk(self, val):
-        logger.debug('MemoryChunkedHandler.new_chunk(sz=%d)', len(val))
-        self.sio.write(val)
-    def end(self):
-        return self.sio.getvalue()
-
+class _TdsLogin:
+    pass
 
 ######################
 ## Connection class ##
 ######################
-class Connection(object):
+class _Connection(object):
     @property
     def as_dict(self):
         """
@@ -107,164 +61,99 @@ class Connection(object):
         Returns current chunk handler
         Default is MemoryChunkedHandler()
         '''
-        return self.tds_socket.chunk_handler
+        if not self._conn:
+            raise Error('Connection closed')
+        return self._conn.chunk_handler
 
     @chunk_handler.setter
     def chunk_handler_set(self, value):
-        self.tds_socket.chunk_handler = value
+        if not self._conn:
+            raise Error('Connection closed')
+        self._conn.chunk_handler = value
 
     @property
     def tds_version(self):
         '''
         Returns version of tds protocol that is being used by this connection
         '''
-        tds = self._get_connection()
-        return tds.tds_version
+        if not self._conn:
+            raise Error('Connection closed')
+        return self._conn.tds_version
 
     @property
     def product_version(self):
         '''
         Returns version of the server
         '''
-        return self.tds_socket.product_version
+        if not self._conn:
+            raise Error('Connection closed')
+        return self._conn.product_version
 
-    def _get_connection(self):
-        if self._closed:
-            raise Error('Connection is closed')
-        if self.tds_socket.is_dead():
-            # clear transaction
-            self.tds_socket.tds72_transaction = '\x00\x00\x00\x00\x00\x00\x00\x00'
-            tds_connect_and_login(self.tds_socket, self._login)
-            self._try_activate_cursor(None)
-            if not self._autocommit:
-                tds_submit_begin_tran(self.tds_socket)
-            self._sqlok()
-        return self.tds_socket
-
-    def __init__(self, server, user, password,
-            charset, database, appname, port, tds_version,
-            as_dict, encryption_level, login_timeout, timeout, autocommit):
-        self._autocommit = autocommit
-        logger.debug("Connection.__init__()")
-        self._charset = ''
+    def __init__(self, login, as_dict, autocommit):
+        self._login = login
         self._as_dict = as_dict
+        self._autocommit = autocommit
         self._state = DB_RES_NO_MORE_RESULTS
-        self.tds_socket = None
-        self._closed = False
         self._active_cursor = None
-
-        # support MS methods of connecting locally
-        instance = ""
-        if "\\" in server:
-            server, instance = server.split("\\")
-
-        if server in (".", "(local)"):
-            server = "localhost"
-
-        self._login = login = tds_alloc_login(1)
-        # set default values for loginrec
-        login.library = "Python TDS Library"
-
-        appname = appname or "pytds"
-
-        login.encryption_level = encryption_level
-        login.user_name = user or ''
-        login.password = password or ''
-        login.app = appname
-        login.port = port
-        if tds_version:
-            login.tds_version = tds_version
-        login.database = database
-
-        # that will set:
-        # ANSI_DEFAULTS to ON,
-        # IMPLICIT_TRANSACTIONS to OFF,
-        # TEXTSIZE to 0x7FFFFFFF (2GB) (TDS 7.2 and below), TEXTSIZE to infinite (introduced in TDS 7.3),
-        # and ROWCOUNT to infinite
-        login.option_flag2 = TDS_ODBC_ON
-
-        # Set the character set name
-        if charset:
-            _charset = charset
-            self._charset = _charset
-            login.charset = self._charset
-
-        # Connect to the server
-        login.server_name = server
-        login.instance_name = instance
-        ctx = _TdsContext()
-        ctx.int_handler = self._int_handler
-        self.tds_socket = tds_alloc_socket(ctx, 512)
-        self.tds_socket.chunk_handler = MemoryChunkedHandler()
-
-        login.connect_timeout = login_timeout
-        login.query_timeout = timeout
-
-    def autocommit(self, status):
-        """
-        Turn autocommit ON or OFF.
-        """
-
-        if status == self._autocommit:
-            return
-
-        self.cancel()
-        if status:
-            tds_submit_rollback(self.tds_socket, False)
-        else:
-            tds_submit_begin_tran(self.tds_socket)
-        self._sqlok()
-        self._autocommit = status
+        from tds import _TdsSocket
+        self._conn = None
+        self._conn = _TdsSocket(self._login)
+        if not self._autocommit:
+            tds_submit_begin_tran(self._conn.main_session)
+        self._sqlok(self._conn.main_session)
 
     def commit(self):
         """
         Commit transaction which is currently in progress.
         """
-
+        if not self._conn:
+            raise Error('Connection closed')
         if self._autocommit:
             return
 
-        tds = self._get_connection()
-        try:
-            self._try_activate_cursor(None)
-            tds_submit_commit(tds, True)
-            self._sqlok()
-            while self._nextset(None):
-                pass
-        except Exception, e:
-            raise OperationalError('Cannot commit transaction: ' + str(e[0]))
+        conn = self._conn
+        if not conn.tds72_transaction:
+            return
+
+        self._try_activate_cursor(None)
+        self._cancel(conn.main_session)
+        tds_submit_commit(conn.main_session, True)
+        self._sqlok(conn.main_session)
+        while self._nextset(conn.main_session):
+            pass
 
     def cursor(self):
         """
         Return cursor object that can be used to make queries and fetch
         results from the database.
         """
-        return Cursor(self)
+        if not self._conn:
+            raise Error('Connection closed')
+        return _Cursor(self)
 
     def rollback(self):
         """
         Roll back transaction which is currently in progress.
         """
+        if not self._conn:
+            raise Error('Connection closed')
         if self._autocommit:
             return
 
-        if not self.tds_socket.is_dead():
-            tds = self._get_connection()
-            self.cancel()
-            self._active_cursor = None
-            tds_submit_rollback(tds, True)
-            self._sqlok()
-            while self._nextset(None):
-                pass
-
-    def _int_handler(self):
-        raise Exception('not implemented')
+        session = self._conn.main_session
+        self._cancel(session)
+        self._active_cursor = None
+        tds_submit_rollback(session, True)
+        self._sqlok(session)
+        while self._nextset(session):
+            pass
 
     def __del__(self):
         logger.debug("MSSQLConnection.__del__()")
-        self.close()
+        if self._conn is not None:
+            self._conn.close()
 
-    def cancel(self):
+    def _cancel(self, session):
         """
         cancel() -- cancel all pending results.
 
@@ -272,11 +161,9 @@ class Connection(object):
         It can be called more than once in a row. No exception is raised in
         this case.
         """
-        logger.debug("MSSQLConnection.cancel()")
-        tds = self.tds_socket
-        if not tds.is_dead():
-            tds_send_cancel(tds)
-            tds_process_cancel(tds)
+        logger.debug("MSSQLConnection._cancel()")
+        tds_send_cancel(session)
+        tds_process_cancel(session)
 
     def close(self):
         """
@@ -287,13 +174,10 @@ class Connection(object):
         this case.
         """
         logger.debug("MSSQLConnection.close()")
-        if self._closed:
+        if not self._conn:
             raise Error('Connection closed')
-        self._closed = True
-        tds = self.tds_socket
-        if tds is not None:
-            tds_close_socket(tds)
-            tds_free_socket(tds)
+        self._conn.close()
+        self._conn = None
 
     def select_db(self, dbname):
         """
@@ -303,16 +187,17 @@ class Connection(object):
         failure.
         """
         logger.debug("MSSQLConnection.select_db()")
+        if not self._conn:
+            raise Error('Connection closed')
         cur = self.cursor()
         try:
-            cur.execute('use {0}'.format(tds_quote_id(self.tds_socket, dbname)))
+            cur.execute('use {0}'.format(tds_quote_id(self._conn, dbname)))
         finally:
             cur.close()
 
-    def _nextrow(self):
+    def _nextrow(self, session):
         logger.debug("_nextrow()")
-        tds = self.tds_socket
-        resinfo = tds.res_info
+        resinfo = session.res_info
         if not resinfo or self._state != DB_RES_RESULTSET_ROWS:
             # no result set or result set empty (no rows)
             logger.debug("leaving _nextrow() returning NO_MORE_ROWS")
@@ -326,16 +211,16 @@ class Connection(object):
         mask = TDS_STOPAT_ROWFMT|TDS_RETURN_DONE|TDS_RETURN_ROW|TDS_RETURN_COMPUTE
 
         # Get the row from the TDS stream.
-        rc, res_type, done_flags = tds_process_tokens(tds, mask)
+        rc, res_type, done_flags = tds_process_tokens(session, mask)
         if done_flags & TDS_DONE_ERROR:
-            raise_db_exception(tds)
+            raise_db_exception(session)
             assert False
             raise Exception('FAIL')
         if rc == TDS_SUCCESS:
             if res_type in (TDS_ROW_RESULT, TDS_COMPUTE_RESULT):
                 # Add the row to the row buffer, whose capacity is always at least 1
-                resinfo = tds.current_results
-                #_, res_type, _ = tds_process_tokens(tds, TDS_TOKEN_TRAILING)
+                resinfo = session.current_results
+                #_, res_type, _ = tds_process_tokens(session, TDS_TOKEN_TRAILING)
             else:
                 self._state = DB_RES_NEXT_RESULT
         elif rc == TDS_NO_MORE_RESULTS:
@@ -343,11 +228,10 @@ class Connection(object):
         else:
             raise Exception("unexpected result from tds_process_tokens")
 
-    def _sqlok(self):
+    def _sqlok(self, session):
         logger.debug("dbsqlok()")
         #CHECK_CONN(FAIL);
 
-        tds = self.tds_socket
         #
         # If we hit an end token -- e.g. if the command
         # submitted returned no data (like an insert) -- then
@@ -355,14 +239,14 @@ class Connection(object):
         #
         logger.debug("dbsqlok() not done, calling tds_process_tokens()")
         while True:
-            tds_code, result_type, done_flags = tds_process_tokens(tds, TDS_TOKEN_RESULTS)
+            tds_code, result_type, done_flags = tds_process_tokens(session, TDS_TOKEN_RESULTS)
 
             #
             # The error flag may be set for any intervening DONEINPROC packet, in particular
             # by a RAISERROR statement.  Microsoft db-lib returns FAIL in that case. 
             #/
             if done_flags & TDS_DONE_ERROR:
-                raise_db_exception(tds)
+                raise_db_exception(session)
                 assert False
                 raise Exception('FAIL')
             if result_type == TDS_ROWFMT_RESULT:
@@ -386,24 +270,24 @@ class Connection(object):
                 logger.error('logic error: tds_process_tokens result_type %d', result_type);
 
     def _fetchone(self, cursor):
-        if self._active_cursor is not cursor:
-            raise Error('This cursor is not active')
         """
         Helper method used by fetchone and fetchmany to fetch and handle
         """
-        tds = self.tds_socket
-        if tds.res_info is None:
+        session = cursor._session
+        if session is None:
+            raise Error('This cursor is not active')
+        if session.res_info is None:
             raise Error("Previous statement didn't produce any results")
 
         if self._state == DB_RES_NO_MORE_RESULTS:
             return None
 
-        self._nextrow()
+        self._nextrow(session)
 
         if self._state != DB_RES_RESULTSET_ROWS:
             return None
 
-        cols = tds.res_info.columns
+        cols = session.res_info.columns
         row = tuple(col.value for col in cols)
         if self.as_dict:
             row_dict = dict(enumerate(cols))
@@ -411,112 +295,166 @@ class Connection(object):
             row = row_dict
         return row
 
-    def _nextset(self, cursor):
-        if cursor is not self._active_cursor:
+    def _nextset(self, session):
+        if session is None:
             raise Error('This cursor is not active')
 
         while self._state == DB_RES_RESULTSET_ROWS:
-            self._nextrow()
-        self._sqlok()
+            self._nextrow(session)
+        self._sqlok(session)
         return None if self._state == DB_RES_NO_MORE_RESULTS else True
 
     def _rowcount(self, cursor):
-        if cursor is not self._active_cursor:
+        session = cursor._session
+        if session is None:
             return -1
-        tds = self.tds_socket
-        return tds.rows_affected
+        return session.rows_affected
 
     def _get_proc_return_status(self, cursor):
-        if cursor is not self._active_cursor:
+        session = cursor._session
+        if session is None:
             return None
-        tds = self.tds_socket
-        if not tds.has_status:
-            tds_process_tokens(tds, TDS_RETURN_PROC)
-        return tds.ret_status if tds.has_status else None
+        if not session.has_status:
+            tds_process_tokens(session, TDS_RETURN_PROC)
+        return session.ret_status if session.has_status else None
 
     def _description(self, cursor):
-        if cursor is not self._active_cursor:
+        session = cursor._session
+        if session is None:
             return None
-        res = self.tds_socket.res_info
+        res = session.res_info
         if res:
             return res.description
         else:
             return None
 
     def _native_description(self):
-        if cursor is not self._active_cursor:
+        session = cursor._session
+        if session is None:
             return None
-        res = self._conn.tds_socket.res_info
+        res = session.res_info
         if res:
             return res.native_descr
         else:
             return None
 
     def _close_cursor(self, cursor):
-        if cursor is self._active_cursor:
-            self._active_cursor = None
+        if cursor._session is not None:
+            if self._conn.mars_enabled:
+                cursor._session.close()
+            else:
+                if cursor is self._active_cursor:
+                    self._active_cursor = None
+                    self._session = None
         cursor._conn = None
 
     def _try_activate_cursor(self, cursor):
-        tds = self._get_connection()
-        if cursor is self._active_cursor or self._active_cursor is None:
-            self.cancel()
-        else:
-            if tds.state == TDS_PENDING:
-                rc, result_type, _ = tds_process_tokens(tds, TDS_TOKEN_TRAILING)
-                if rc != TDS_NO_MORE_RESULTS:
-                    raise InterfaceError('Results are still pending on connection')
-        self._active_cursor = cursor
+        conn = self._conn
+        if not conn.mars_enabled:
+            if not (cursor is self._active_cursor or self._active_cursor is None):
+                session = conn.main_session
+                if session.state == TDS_PENDING:
+                    rc, result_type, _ = tds_process_tokens(session, TDS_TOKEN_TRAILING)
+                    if rc != TDS_NO_MORE_RESULTS:
+                        raise InterfaceError('Results are still pending on connection')
+                cursor._session = session
+            self._active_cursor = cursor
 
     def _execute(self, cursor, operation, params):
-        tds = self._get_connection()
+        if not self._conn:
+            raise Error('Connection closed')
+        tds = self._conn
         self._try_activate_cursor(cursor)
-        tds_submit_query(tds, operation, params)
-        self._state = DB_RES_INIT
-        self._sqlok()
-
-    def _callproc(self, cursor, procname, parameters):
-        logger.debug('callproc begin')
-        tds = self._get_connection()
-        self._try_activate_cursor(cursor)
-        tds_submit_rpc(tds, procname, parameters)
-        tds.output_params = {}
+        session = cursor._session
+        self._cancel(session)
+        tds_submit_query(session, operation, params)
         self._state = DB_RES_INIT
         while True:
-            tds_code, result_type, done_flags = tds_process_tokens(tds, TDS_TOKEN_RESULTS)
+            tds_code, result_type, done_flags = tds_process_tokens(session, TDS_TOKEN_RESULTS)
+
             #
             # The error flag may be set for any intervening DONEINPROC packet, in particular
             # by a RAISERROR statement.  Microsoft db-lib returns FAIL in that case. 
             #/
             if done_flags & TDS_DONE_ERROR:
-                raise_db_exception(tds)
+                raise_db_exception(session)
                 assert False
                 raise Exception('FAIL')
-            if result_type == TDS_ROWFMT_RESULT:
+            if result_type == TDS_STATUS_RESULT:
+                continue
+            elif result_type == TDS_DONEINPROC_RESULT:
+                if not done_flags & TDS_DONE_COUNT:
+                    # skip results that don't event have rowcount
+                    continue
+                self._state = DB_RES_RESULTSET_EMPTY
+                break
+            elif result_type == TDS_ROWFMT_RESULT:
                 self._state = DB_RES_RESULTSET_ROWS
                 break
-            elif result_type == TDS_DONEINPROC_RESULT:
-                self._state = DB_RES_RESULTSET_EMPTY
             elif result_type in (TDS_DONE_RESULT, TDS_DONEPROC_RESULT):
                 if done_flags & TDS_DONE_MORE_RESULTS:
+                    if not done_flags & TDS_DONE_COUNT:
+                        # skip results that don't event have rowcount
+                        continue
                     self._state = DB_RES_NEXT_RESULT
                 else:
                     self._state = DB_RES_NO_MORE_RESULTS
                 break
-            elif result_type == TDS_STATUS_RESULT:
+            else:
+                logger.error('logic error: tds_process_tokens result_type %d', result_type);
+
+    def _callproc(self, cursor, procname, parameters):
+        logger.debug('callproc begin')
+        if not self._conn:
+            raise Error('Connection closed')
+        tds = self._conn
+        self._try_activate_cursor(cursor)
+        session = cursor._session
+        self._cancel(session)
+        tds_submit_rpc(session, procname, parameters)
+        session.output_params = {}
+        self._state = DB_RES_INIT
+        while True:
+            tds_code, result_type, done_flags = tds_process_tokens(session, TDS_TOKEN_RESULTS)
+            #
+            # The error flag may be set for any intervening DONEINPROC packet, in particular
+            # by a RAISERROR statement.  Microsoft db-lib returns FAIL in that case. 
+            #/
+            if done_flags & TDS_DONE_ERROR:
+                raise_db_exception(session)
+                assert False
+                raise Exception('FAIL')
+            if result_type == TDS_STATUS_RESULT:
                 continue
+            elif result_type == TDS_PARAM_RESULT:
+                continue
+            elif result_type == TDS_DONEINPROC_RESULT:
+                self._state = DB_RES_RESULTSET_EMPTY
+                continue
+            elif result_type == TDS_ROWFMT_RESULT:
+                self._state = DB_RES_RESULTSET_ROWS
+                break
+            elif result_type in (TDS_DONE_RESULT, TDS_DONEPROC_RESULT):
+                if done_flags & TDS_DONE_MORE_RESULTS:
+                    if not done_flags & TDS_DONE_COUNT:
+                        # skip results that don't event have rowcount
+                        continue
+                    self._state = DB_RES_NEXT_RESULT
+                else:
+                    self._state = DB_RES_NO_MORE_RESULTS
+                break
             else:
                 logger.error('logic error: tds_process_tokens result_type %d', result_type);
         logger.debug('callproc end')
         results = list(parameters)
-        for key, param in tds.output_params.items():
+        for key, param in session.output_params.items():
             results[key] = param.value
         return results
 
 ##################
 ## Cursor class ##
 ##################
-class Cursor(object):
+class _Cursor(object):
     """
     This class represents a database cursor, which is used to issue queries
     and fetch results from a database connection.
@@ -525,6 +463,10 @@ class Cursor(object):
         self._conn = conn
         self._batchsize = 1
         self.arraysize = 1
+        if conn._conn.mars_enabled:
+            self._session = conn._conn.create_session()
+        else:
+            self._session = conn._conn.main_session
 
     def __enter__(self):
         return self
@@ -555,6 +497,10 @@ class Cursor(object):
     def return_value(self):
         return self.get_proc_return_status()
 
+    @property
+    def connection(self):
+        return self._conn
+
     def get_proc_return_status(self):
         return self._conn._get_proc_return_status(self)
 
@@ -585,13 +531,12 @@ class Cursor(object):
 
     def executemany(self, operation, params_seq):
         counts = []
-        tds = self._conn.tds_socket
         for params in params_seq:
             self.execute(operation, params)
-            if tds.rows_affected != -1:
-                counts.append(tds.rows_affected)
+            if self._session.rows_affected != -1:
+                counts.append(self._session.rows_affected)
         if counts:
-            tds.rows_affected = sum(counts)
+            self._session.rows_affected = sum(counts)
 
     def execute_scalar(self, query_string, params=None):
         """
@@ -621,7 +566,7 @@ class Cursor(object):
         return row[0]
 
     def nextset(self):
-        return self._conn._nextset(self)
+        return self._conn._nextset(self._session)
 
     @property
     def rowcount(self):
@@ -672,9 +617,10 @@ class Cursor(object):
         pass
 
 def connect(server='.', database='', user='', password='', timeout=0,
-        login_timeout=60, charset=None, as_dict=False,
+        login_timeout=60, as_dict=False,
         host='', appname=None, port=None, tds_version=TDS74,
-        encryption_level=TDS_ENCRYPTION_OFF, autocommit=True):
+        encryption_level=TDS_ENCRYPTION_OFF, autocommit=True,
+        blocksize=4096, use_mars=True):
     """
     Constructor for creating a connection to the database. Returns a
     Connection object.
@@ -691,8 +637,6 @@ def connect(server='.', database='', user='', password='', timeout=0,
     :type timeout: int
     :param login_timeout: timeout for connection and login in seconds, default 60
     :type login_timeout: int
-    :param charset: character set with which to connect to the database
-    :type charset: string
     :keyword as_dict: whether rows should be returned as dictionaries instead of tuples.
     :type as_dict: boolean
     :keyword appname: Set the application name to use for the connection
@@ -716,11 +660,44 @@ def connect(server='.', database='', user='', password='', timeout=0,
     if host:
         server = host
 
-    conn = Connection(server, user, password, charset, database,
-        appname, port, tds_version=tds_version, as_dict=as_dict, login_timeout=login_timeout,
-        timeout=timeout, encryption_level=encryption_level, autocommit=autocommit)
+    # support MS methods of connecting locally
+    instance = ""
+    if "\\" in server:
+        server, instance = server.split("\\")
 
-    return conn
+    if server in (".", "(local)"):
+        server = "localhost"
+
+    login = _TdsLogin()
+    login.library = "Python TDS Library"
+    login.encryption_level = encryption_level
+    login.user_name = user or ''
+    login.password = password or ''
+    login.app_name = appname or 'pytds'
+    login.port = port
+    login.language = '' # use database default
+    login.attach_db_file = ''
+    login.tds_version = tds_version
+    login.database = database
+    login.emul_little_endian = False
+    login.bulk_copy = False
+    login.text_size = 0
+    login.client_lcid = lcid.LANGID_ENGLISH_US
+    login.use_mars = use_mars
+
+    # that will set:
+    # ANSI_DEFAULTS to ON,
+    # IMPLICIT_TRANSACTIONS to OFF,
+    # TEXTSIZE to 0x7FFFFFFF (2GB) (TDS 7.2 and below), TEXTSIZE to infinite (introduced in TDS 7.3),
+    # and ROWCOUNT to infinite
+    login.option_flag2 = TDS_ODBC_ON
+
+    login.connect_timeout = login_timeout
+    login.query_timeout = timeout
+    login.server_name = server
+    login.instance_name = instance
+    login.blocksize = blocksize
+    return _Connection(login, as_dict, autocommit)
 
 def Date(year, month, day):
     return date(year, month, day)
@@ -736,8 +713,8 @@ def TimeFromTicks(ticks):
     import time
     return Time(*time.localtime(ticks)[3:6])
 
-def Timestamp(year, month, day, hour, minute, second):
-    return datetime(year, month, day, hour, minute, second)
+def Timestamp(year, month, day, hour, minute, second, microseconds=0, tzinfo=None):
+    return datetime(year, month, day, hour, minute, second, microseconds, tzinfo)
 
 def TimestampFromTicks(ticks):
     return datetime.fromtimestamp(ticks)
