@@ -1,4 +1,5 @@
 # vim: set fileencoding=utf8 :
+import sys
 import struct
 import os
 import logging
@@ -12,60 +13,98 @@ else:
 from tdsproto import *
 from tds import *
 from token import *
+from smp import SmpManager
 
 logger = logging.getLogger(__name__)
 
+class _SspiAuthentication(object):
+    def __init__(self, login):
+        import sspi
+        import ctypes
+        # parse username/password informations
+        user_name = login.user_name
+        pos = user_name.find('\\')
+        identity = None
+        if pos != -1:
+            domain = user_name[:pos]
+            user_name = user_name[pos+1:]
+            identity = sspi.make_winnt_identity(
+                domain,
+                user_name,
+                login.password)
 
-def tds_connect_and_login(tds, login):
-    tds.login = login
-    tds.tds_version = login.tds_version
-    tds_conn(tds).emul_little_endian = login.emul_little_endian
-    if IS_TDS7_PLUS(tds):
-        # TDS 7/8 only supports little endian
-        tds_conn(tds).emul_little_endian = True
-    if IS_TDS7_PLUS(tds) and login.instance_name and not login.port:
-        instances = tds7_get_instances(login.ip_addr or login.server_name)
-        if login.instance_name not in instances:
-            raise LoginError("Instance {0} not found on server {1}".format(login.instance_name, login.server_name))
-        instdict = instances[login.instance_name]
-        if 'tcp' not in instdict:
-            raise LoginError("Instance {0} doen't have tcp connections enabled".format(login.instance_name))
-        login.port = int(instdict['tcp'])
-    connect_timeout = login.connect_timeout
-    tds.query_timeout = connect_timeout if connect_timeout else login.query_timeout
-    try:
-        tds_open_socket(tds, login.ip_addr or login.server_name, login.port, connect_timeout)
-    except socket.error as e:
-        raise LoginError("Cannot connect to server '{0}': {1}".format(login.server_name, e), e)
-    tds_set_state(tds, TDS_IDLE)
-    try:
-        db_selected = False
-        if IS_TDS71_PLUS(tds):
-            tds71_do_login(tds, login)
-            db_selected = True
-        elif IS_TDS7_PLUS(tds):
-            tds7_send_login(tds, login)
-            db_selected = True
-        else:
-            raise NotImplementedError('This TDS version is not supported')
-            tds._writer.begin_packet(TDS_LOGIN)
-            tds_send_login(tds, login)
-        if not tds_process_login_tokens(tds):
-            raise_db_exception(tds)
-            #raise LoginError("Cannot connect to server '{0}' as user '{1}'".format(login.server_name, login.user_name))
-        text_size = login.text_size
-        if text_size or not db_selected and login.database:
-            q = []
-            if text_size:
-                q.append('set textsize {0}'.format(int(text_size)))
-            if not db_selected and login.database:
-                q.append('use ' + tds_quote_id(tds, login.database))
-            tds_submit_query(tds, ''.join(q))
-            tds_process_simple_query(tds)
-        return tds
-    except:
-        tds_close_socket(tds)
-        raise
+        # using Negotiate system will use proper protocol (either NTLM or Kerberos)
+        self._cred = sspi.SspiCredentials(
+            package='Negotiate',
+            use=sspi.SECPKG_CRED_OUTBOUND,
+            identity=identity)
+
+        # build SPN
+        primary_host_name, _, _ = socket.gethostbyname_ex(login.server_name)
+        self._sname = 'MSSQLSvc/{0}:{1}'.format(primary_host_name, login.port)
+        self._buf = ctypes.create_string_buffer(4096)
+        bufs = [(sspi.SECBUFFER_TOKEN, self._buf)]
+        self._flags = sspi.ISC_REQ_CONFIDENTIALITY|sspi.ISC_REQ_REPLAY_DETECT|sspi.ISC_REQ_CONNECTION
+        self._ctx, status, bufs = self._cred.create_context(
+            flags=self._flags,
+            byte_ordering='network',
+            target_name=self._sname,
+            output_buffers=bufs)
+        if status == sspi.Status.SEC_I_COMPLETE_AND_CONTINUE or status == sspi.Status.SEC_I_CONTINUE_NEEDED:
+            self._ctx.complete_auth_token(bufs)
+        self.packet = bufs[0][1]
+
+    def handle_next(self, tds, length):
+        import sspi
+        r = tds._reader
+        token = r.readall(length)
+        status, buffers = self._ctx.next(
+            flags=self._flags,
+            byte_ordering='network',
+            target_name=self._sname,
+            input_buffers=[(sspi.SECBUFFER_TOKEN, token)],
+            output_buffers=[(sspi.SECBUFFER_TOKEN, self._buf)])
+        if len(buffers[0][1]) == 0:
+            return
+        w = tds._writer
+        w.write(buffers[0][1])
+        w.flush()
+    
+    def close(self):
+        self._ctx.close()
+        self._cred.close()
+
+class _NtlmAuth(object):
+    def __init__(self, login):
+        self._login = login
+        import ntlm
+        self._domain, self._user = login.user_name.split('\\', 1)
+        self.packet = ntlm.create_NTLM_NEGOTIATE_MESSAGE_raw(login.client_host_name, self._domain)
+    def close(self):
+        pass
+    def handle_next(self, tds, length):
+        import ntlm
+        packet = tds._reader.readall(length)
+        nonce, flags = ntlm.parse_NTLM_CHALLENGE_MESSAGE_raw(packet)
+        w = tds._writer
+        w.write(ntlm.create_NTLM_AUTHENTICATE_MESSAGE_raw(nonce, self._user, self._domain, self._login.password, flags))
+        w.flush()
+
+def tds_login(tds, login):
+    db_selected = False
+    if IS_TDS71_PLUS(tds):
+        tds71_do_login(tds, login)
+        db_selected = True
+    elif IS_TDS7_PLUS(tds):
+        tds7_send_login(tds, login)
+        db_selected = True
+    else:
+        raise NotImplementedError('This TDS version is not supported')
+        tds._writer.begin_packet(TDS_LOGIN)
+        tds_send_login(tds, login)
+    if not tds_process_login_tokens(tds):
+        raise_db_exception(tds)
+        #raise LoginError("Cannot connect to server '{0}' as user '{1}'".format(login.server_name, login.user_name))
 
 import socket
 this_host_name = socket.gethostname()
@@ -84,15 +123,21 @@ def tds7_send_login(tds, login):
         raise Error('Password should be not more than 128 characters')
     current_pos = 86 + 8 if IS_TDS72_PLUS(tds) else 86
     client_host_name = this_host_name
+    login.client_host_name = client_host_name
     packet_size = current_pos + (len(client_host_name) + len(login.app_name) + len(login.server_name) + len(login.library) + len(login.language) + len(login.database))*2
     auth_len = 0
-    # TODO: support sspi login
-    if False:
+    if sys.platform == 'win32':
         if user_name.find('\\') != -1 or not user_name:
-            raise NotImplementedError('sspi not implemented')
+            tds.authentication = _SspiAuthentication(login)
+            auth_len = len(tds.authentication.packet)
+            packet_size += auth_len
+        else:
+            packet_size += (len(user_name) + len(login.password))*2
     else:
         if user_name.find('\\') != -1:
-            raise NotImplementedError('ntlm not implemented')
+            tds.authentication = _NtlmAuth(login)
+            auth_len = len(tds.authentication.packet)
+            packet_size += auth_len
         elif not user_name:
             raise NotImplementedError('requested GSS authentication but it is not implemented')
         else:
@@ -108,9 +153,8 @@ def tds7_send_login(tds, login):
     if not login.bulk_copy:
         option_flag1 |= TDS_DUMPLOAD_OFF
     w.put_byte(option_flag1)
-    if False:
-        if tds.authentication:
-            option_flag2 |= TDS_INTEGRATED_SECURITY_ON
+    if tds.authentication:
+        option_flag2 |= TDS_INTEGRATED_SECURITY_ON
     w.put_byte(option_flag2)
     w.put_byte(0) # sql_type_flag
     option_flag3 = TDS_UNKNOWN_COLLATION_HANDLING
@@ -251,7 +295,7 @@ def tds71_do_login(tds, login):
     w.put_int(os.getpid()) # TODO: change this to thread id
     if IS_TDS72_PLUS(tds):
         # MARS (1 enabled)
-        w.put_byte(0)
+        w.put_byte(1 if login.use_mars else 0)
     w.flush()
     p = tds._reader.read_whole_packet()
     size = len(p)
@@ -275,11 +319,11 @@ def tds71_do_login(tds, login):
         if off > size or off + l > size:
             raise Error('TDS_FAIL')
         if type == VERSION:
-            tds.product_version = prod_version_struct.unpack_from(p, off)
+            tds.conn.product_version = prod_version_struct.unpack_from(p, off)
         elif type == ENCRYPTION and l >= 1:
             crypt_flag, = byte_struct.unpack_from(p, off)
         elif type == MARS:
-            tds.mars_enabled = bool(byte_struct.unpack_from(p, off)[0])
+            tds.conn._mars_enabled = bool(byte_struct.unpack_from(p, off)[0])
         i += 5
     # we readed all packet
     logger.debug('detected flag %d', crypt_flag)
