@@ -138,7 +138,7 @@ def is_ascii_type(x): return x in (XSYBCHAR,XSYBVARCHAR,SYBTEXT,SYBCHAR,SYBVARCH
 def is_char_type(x): return is_unicode_type(x) or is_ascii_type(x)
 def is_similar_type(x, y): return is_char_type(x) and is_char_type(y) or is_unicode_type(x) and is_unicode_type(y)
 
-def tds_conn(tds): return tds.conn
+def tds_conn(tds): return tds
 
 def TDS_IS_SOCKET_INVALID(sock):
     return sock is None
@@ -164,74 +164,6 @@ to_client = 1
 TDS_DATETIME = struct.Struct('<ll')
 TDS_DATETIME4 = struct.Struct('<HH')
 
-#
-# Convert from db date format to a structured date format
-# @param datetype source date type. SYBDATETIME or SYBDATETIME4
-# @param di       source date
-# @param dr       destination date
-# @return TDS_FAIL or TDS_SUCCESS
-#
-def tds_datecrack(datetype, di):
-    if datetype == (SYBMSDATE, SYBMSTIME, SYBMSDATETIME2, SYBMSDATETIMEOFFSET):
-        # I think this is not a real wire format
-        raise Exception('not implemented')
-        #const TDS_DATETIMEALL *dta = (const TDS_DATETIMEALL *) di;
-        #dt_days = (datetype == SYBMSTIME) ? 0 : dta->date;
-        #if (datetype == SYBMSDATE) {
-        #    dms = 0;
-        #    secs = 0;
-        #    dt_time = 0;
-        #} else {
-        #    dms = dta->time % 10000000u;
-        #    dt_time = dta->time / 10000000u;
-        #    secs = dt_time % 60;
-        #    dt_time = dt_time / 60;
-        #}
-        #if (datetype == SYBMSDATETIMEOFFSET) {
-        #    --dt_days;
-        #    dt_time = dt_time + 86400 + dta->offset;
-        #    dt_days += dt_time / 86400;
-        #    dt_time %= 86400;
-        #}
-    elif datetype == SYBDATETIME or datetype == SYBDATETIMN and len(di) == 8:
-        dt_days, dt_time = TDS_DATETIME.unpack(di)
-        dms = ((dt_time % 300) * 1000 + 150) / 300 * 10000
-        dt_time = dt_time / 300
-        secs = dt_time % 60
-        dt_time = dt_time / 60
-    elif datetype == SYBDATETIME4 or datetype == SYBDATETIMN and len(di) == 4:
-        dt_days, dt_time = TDS_DATETIME4.unpack(di)
-        secs = 0;
-        dms = 0;
-    else:
-        raise Exception('TDS_FAIL')
-
-    #
-    # -53690 is minimun  (1753-1-1) (Gregorian calendar start in 1732) 
-    # 2958463 is maximun (9999-12-31)
-    #
-    l = dt_days + (146038 + 146097*4)
-    #wday = (l + 4) % 7
-    n = (4 * l) / 146097 # n century
-    l = l - (146097 * n + 3) / 4 # days from xx00-02-28 (y-m-d)
-    i = (4000 * (l + 1)) / 1461001 # years from xx00-02-28
-    l = l - (1461 * i) / 4 # year days from xx00-02-28
-    #ydays = l - 305 if l >= 306 else l + 60
-    l += 31
-    j = (80 * l) / 2447
-    days = l - (2447 * j) / 80
-    l = j / 11
-    months = j + 1 - 12 * l
-    years = 100 * (n - 1) + i + l
-    #if l == 0 and (years & 3) == 0 and (years % 100 != 0 or years % 400 == 0):
-    #    ++ydays
-
-    hours = dt_time / 60
-    mins = dt_time % 60
-
-    from datetime import datetime
-
-    return datetime(years, months + 1, days, hours, mins, secs, dms/10)
 
 #
 # Quote an id
@@ -421,7 +353,6 @@ def tds_set_state(tds, state):
             tds_mutex_unlock(tds.wire_mtx)
         tds.state = state
     elif state == TDS_QUERYING:
-        #check_tds_extra(tds)
         if tds_mutex_trylock(tds.wire_mtx):
             return tds.state
         if tds.state == TDS_DEAD:
@@ -433,14 +364,11 @@ def tds_set_state(tds, state):
             logger.error('logic error: cannot change query state from {0} to {1}'.\
                     format(state_names[prior_state], state_names[state]))
         else:
-            #tds_free_all_results(tds)
             tds.rows_affected = TDS_NO_COUNT
-            #tds_release_cursor
             tds.internal_sp_called = 0
             tds.state = state
     else:
         assert False
-    #check_tds_extra(tds)
     return tds.state
 
 def tds_mutex_trylock(mutex):
@@ -484,6 +412,7 @@ class _TdsReader(object):
         self._have = 0 # number of bytes read from packet
         self._size = 0 # size of current packet
         self._tds = tds
+        self._transport = tds
         self._type = None
         self._status = None
 
@@ -575,7 +504,7 @@ class _TdsReader(object):
             if self._have >= self._size:
                 self._read_packet()
             else:
-                self._buf = self._tds._read(self._size - self._have)
+                self._buf = self._transport.recv(self._size - self._have)
                 self._pos = 0
                 self._have += len(self._buf)
         res = self._buf[self._pos:self._pos+size]
@@ -583,19 +512,17 @@ class _TdsReader(object):
         return res
 
     def _read_packet(self):
-        if self._tds.is_dead():
-            raise Exception('Read attempt when state is TDS_DEAD')
-        header = self._tds._read(_header.size)
+        header = self._transport.recv(_header.size)
         if len(header) < _header.size:
             self._pos = 0
             if self._tds.state != TDS_IDLE and len(header) == 0:
-                tds_close_socket(self._tds)
+                self._transport.close()
             raise Exception('Reading header error')
         logger.debug('Received header')
         self._type, self._status, self._size, self._spid = _header.unpack(header)
         self._have = _header.size
         assert self._size > self._have, 'Empty packet doesn make any sense'
-        self._buf = self._tds._read(self._size - self._have)
+        self._buf = self._transport.recv(self._size - self._have)
         self._have += len(self._buf)
         self._pos = 0
 
@@ -606,6 +533,7 @@ class _TdsReader(object):
 class _TdsWriter(object):
     def __init__(self, tds, bufsize):
         self._tds = tds
+        self._transport = tds
         self._pos = 0
         self._buf = bytearray(bufsize)
 
@@ -711,48 +639,135 @@ class _TdsWriter(object):
         _header.pack_into(self._buf, 0, self._type, status, self._pos, 0)
         if IS_TDS7_PLUS(self._tds) and not self._tds.login:
             self._buf[6] = 0x01
-        self._tds._write(self._buf[:self._pos], final)
+        self._transport.send(self._buf[:self._pos], final)
         self._pos = 8
 
-class _TdsSocket(object):
-    def __init__(self, context, bufsize):
-        self.conn = _TdsConn()
+class MemoryChunkedHandler(object):
+    def begin(self, column, size):
+        logger.debug('MemoryChunkedHandler.begin(sz=%d)', size)
+        self.size = size
+        self.sio = StringIO()
+    def new_chunk(self, val):
+        logger.debug('MemoryChunkedHandler.new_chunk(sz=%d)', len(val))
+        self.sio.write(val)
+    def end(self):
+        return self.sio.getvalue()
+
+class _TdsSession(object):
+    def __init__(self, tds, transport):
         self.out_pos = 8
-        self.login = None
-        self.int_handler = None
-        self.msg_handler = None
         self.res_info = None
         self.in_cancel = False
-        self.env = _TdsEnv()
         self.wire_mtx = None
         self.current_results = None
         self.param_info = None
         self.cur_cursor = None
-        self.collation = None
-        self.tds72_transaction = '\x00\x00\x00\x00\x00\x00\x00\x00'
         self.has_status = False
-        self.messages = []
-        self._reader = _TdsReader(self)
-        self._writer = _TdsWriter(self, bufsize)
-        self.conn.tds_ctx = context
+        self._transport = transport
+        self._reader = _TdsReader(tds)
+        self._reader._transport = transport
+        self._writer = _TdsWriter(tds, tds._bufsize)
+        self._writer._transport = transport
         self.in_buf_max = 0
-        tds_conn(self).s_signal = tds_conn(self).s_signaled = None
-
-        # Jeff's hack, init to no timeout
-        self.query_timeout = 0
-        tds_set_s(self, None)
-        import socket
-        if hasattr(socket, 'socketpair'):
-            tds_conn(self).s_signal, tds_conn(self).s_signaled = socket.socketpair(socket.AF_UNIX, socket.SOCK_DGRAM)
-        self.state = TDS_DEAD
+        self.state = TDS_IDLE
         self.write_mtx = TDS_MUTEX_INIT(self.wire_mtx)
+        self._tds = tds
+        self.messages = []
+        self.chunk_handler = tds.chunk_handler
+        self.rows_affected = -1
 
     def is_dead(self):
         return self.state == TDS_DEAD
 
+    @property
+    def tds_version(self):
+        return self._tds.tds_version
+
+    @property
+    def conn(self):
+        return self._tds
+
+    def close(self):
+        self._transport.close()
+
+class _TdsSocket(object):
+    def __init__(self, login):
+        self._bufsize = login.blocksize
+        self.login = None
+        self.int_handler = None
+        self.msg_handler = None
+        self.env = _TdsEnv()
+        self.collation = None
+        self.tds72_transaction = None
+        self.authentication = None
+        self._mars_enabled = False
+        tds_conn(self).s_signal = tds_conn(self).s_signaled = None
+        self.emul_little_endian = True
+        self.chunk_handler = MemoryChunkedHandler()
+        self._main_session = _TdsSession(self, self)
+
+        # Jeff's hack, init to no timeout
+        self.query_timeout = login.connect_timeout if login.connect_timeout else login.query_timeout
+        tds_set_s(self, None)
+        import socket
+        if hasattr(socket, 'socketpair'):
+            tds_conn(self).s_signal, tds_conn(self).s_signaled = socket.socketpair(socket.AF_UNIX, socket.SOCK_DGRAM)
+        self._login = login
+        self.tds_version = tds_version = login.tds_version
+        self.emul_little_endian = login.emul_little_endian
+        if IS_TDS7_PLUS(self):
+            # TDS 7/8 only supports little endian
+            self.emul_little_endian = True
+        if IS_TDS7_PLUS(self) and login.instance_name and not login.port:
+            instances = tds7_get_instances(login.server_name)
+            if login.instance_name not in instances:
+                raise LoginError("Instance {0} not found on server {1}".format(login.instance_name, login.server_name))
+            instdict = instances[login.instance_name]
+            if 'tcp' not in instdict:
+                raise LoginError("Instance {0} doen't have tcp connections enabled".format(login.instance_name))
+            login.port = int(instdict['tcp'])
+        connect_timeout = login.connect_timeout
+
+        if not login.port:
+            login.port = 1433
+        try:
+            tds_open_socket(self, login.server_name, login.port, connect_timeout)
+        except socket.error as e:
+            raise LoginError("Cannot connect to server '{0}': {1}".format(login.server_name, e), e)
+        try:
+            from login import tds_login
+            tds_login(self._main_session, login)
+            text_size = login.text_size
+            if self.mars_enabled:
+                self._setup_smp()
+            q = []
+            if text_size:
+                q.append('set textsize {0}'.format(int(text_size)))
+            if login.database and self.env.database != login.database:
+                q.append('use ' + tds_quote_id(self, login.database))
+            if q:
+                tds_submit_query(tds._main_session, ''.join(q))
+                tds_process_simple_query(tds._main_session)
+        except:
+            self.close()
+            raise
+
+    def _setup_smp(self):
+        from smp import SmpManager
+        self._smp_manager = SmpManager(self)
+        self._main_session = _TdsSession(self, self._smp_manager.create_session())
+
+    @property
+    def mars_enabled(self): return self._mars_enabled
+
+    @property
+    def main_session(self):
+        return self._main_session
+
+    def create_session(self):
+        return _TdsSession(self, self._smp_manager.create_session())
+
     def _read(self, size):
-        if self.is_dead():
-            raise Exception('Tds is dead')
         events = tds_select(self, TDSSELREAD, self.query_timeout)
         if events & TDSPOLLURG:
             buf = tds_conn(self).s_signaled.read(size)
@@ -762,13 +777,19 @@ class _TdsSocket(object):
         elif events:
             buf = self._sock.recv(size)
             if len(buf) == 0:
-                tds_close_socket(self)
+                self.close()
                 raise Error('Server closed connection')
             return buf
         else:
-            tds_close_socket(self)
+            self.close()
             raise Error('Timeout')
 
+    def recv(self, size):
+        return self._read(size)
+
+    def send(self, data, final):
+        return self._write(data, final)
+
     def _write(self, data, final):
         pos = 0
         while pos < len(data):
@@ -786,59 +807,28 @@ class _TdsSocket(object):
                 nput = self._sock.send(data[pos:], flags)
             except socket.error as e:
                 if e.errno != errno.EWOULDBLOCK:
-                    tds_close_socket(self)
+                    self.close()
                     raise
             pos += nput
         if final and USE_CORK:
             self._sock.setsockopt(socket.SOL_TCP, socket.TCP_CORK, 0)
             self._sock.setsockopt(socket.SOL_TCP, socket.TCP_CORK, 1)
 
-
-def tds_alloc_socket(context, bufsize):
-    return _TdsSocket(context, bufsize)
-
-def tds_free_socket(tds):
-    if tds:
-        #if tds_conn(tds).authentication:
-        #    tds_conn(tds).authentication.free(tds, tds_conn(tds).authentication)
-        #tds_conn(tds).authentication = None
-        #tds_free_all_results(tds)
-        #tds_free_env(tds)
-        #while (tds->dyns)
-        #    tds_free_dynamic(tds, tds->dyns);
-        #while (tds->cursors)
-        #    tds_cursor_deallocated(tds, tds->cursors);
-        #free(tds->in_buf)
-        tds_ssl_deinit(tds)
-        tds_close_socket(tds);
-        if tds_conn(tds).s_signal is not None:
-            tds_conn(tds).s_signal.close()
-        if tds_conn(tds).s_signaled is not None:
-            tds_conn(tds).s_signaled.close()
-        #free(tds_conn(tds)->product_name);
-        #free(tds);
-
-def tds_free_all_results(tds):
-    logger.debug("tds_free_all_results()")
-    if tds.current_results is tds.res_info:
-        tds.current_results = None
-    tds.res_info = None
-    if tds.current_results is tds.param_info:
-        tds.current_results = None
-    tds.param_info = None
-    tds.has_status = 0
-    tds.ret_status = 0
-
-class _OnServer(object):
-    def __init__(self):
-        self.column_type = None
+    def close(self):
+        self._sock.close()
+        tds_set_state(self._main_session, TDS_DEAD)
+        if self.authentication:
+            self.authentication.close()
+            self.authentication = None
+        #tds_ssl_deinit(self)
+        if tds_conn(self).s_signal is not None:
+            tds_conn(self).s_signal.close()
+        if tds_conn(self).s_signaled is not None:
+            tds_conn(self).s_signaled.close()
 
 class _Column(object):
     def __init__(self):
-        self.on_server = _OnServer()
         self.char_codec = None
-        self.column_nullbind = None
-        self.column_varaddr = 0
         self.column_name = ''
         self.value = None
 
@@ -846,69 +836,11 @@ class _Column(object):
         return '<_Column(name={0}), value={1}>'.format(self.column_name, repr(self.value))
 
 class _Results(object):
-    pass
-
-def tds_alloc_results(num_cols):
-    res_info = _Results()
-    res_info.ref_count = 1
-    res_info.columns = []
-    for col in range(num_cols):
-        res_info.columns.append(tds_alloc_column())
-    res_info.num_cols = num_cols
-    res_info.row_size = 0
-    res_info.row_count = 0
-    return res_info
-
-def tds_alloc_column():
-    return _Column()
-
-#
-# Allocate space for row store
-# return NULL on out of memory
-#
-def tds_alloc_row(res_info):
-    # compute row size
-    res_info.row_size = len(res_info.columns)
-
-    res_info.current_row = []
-
-class _TdsLogin:
     def __init__(self):
-        self.option_flag2 = 0
-        self.tds_version = None
-        self.emul_little_endian = False
-        self.port = 1433
-        self.block_size = 4096
-        self.bulk_copy = False
-        self.text_size = 0
-        self.encryption_level = 0
-        self.client_lcid = lcid.LANGID_ENGLISH_US
-
-def tds_alloc_login(use_environment):
-    login = _TdsLogin()
-    login.server_name = ''
-    login.language = '' # if empty use database default
-    login.server_charset = ''
-    login.server_host_name = ''
-    login.app_name = ''
-    login.user_name = ''
-    login.password = ''
-    login.library = 'python-tds'
-    login.ip_addr = ''
-    login.database = ''
-    login.dump_file = ''
-    login.client_charset = ''
-    login.instance_name = ''
-    login.server_realm_name = ''
-    login.attach_db_file = ''
-    return login
-
-class _TdsContext:
-    pass
+        self.columns = []
+        self.row_count = 0
 
 def tds_open_socket(tds, host, port, timeout=0):
-    if not port:
-        port = 1433
     #tds = _Tds(socket.socket(socket.AF_INET, socket.SOCK_STREAM, 0))
     #tds._sock.setsockopt(socket.SOL_SOCKET, socket.SO_KEEPALIVE, struct.pack('i', 1))
     #tds._sock.setsockopt(socket.SOL_TCP, socket.TCP_CORK, struct.pack('i', 1))
@@ -928,13 +860,9 @@ def tds_open_socket(tds, host, port, timeout=0):
     if not timeout:
         timeout = 90000
     tds._sock = socket.create_connection((host, port), timeout)
+    tds._sock.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
     return tds
 
-def tds_close_socket(tds):
-    if not tds.is_dead():
-        tds._sock.close()
-        tds_set_state(tds, TDS_DEAD)
-
 def tds_select(tds, tds_sel, timeout_seconds):
     poll_seconds = 1 if tds.int_handler else timeout_seconds
     seconds = timeout_seconds
