@@ -48,6 +48,13 @@ token_names = {
 def tds_token_name(marker):
     return token_names.get(marker, '')
 
+def tds_process_auth(tds):
+    r = tds._reader
+    pdu_size = r.get_smallint()
+    if not tds.authentication:
+        raise Error('Got unexpected token')
+    tds.authentication.handle_next(tds, pdu_size)
+
 def tds_process_default_tokens(tds, marker):
     r = tds._reader
     logger.debug('tds_process_default_tokens() marker is {0:x}({1})'.format(marker, tds_token_name(marker)))
@@ -147,12 +154,12 @@ def tds_process_row(tds):
     if not info:
         raise Exception('TDS_FAIL')
 
-    assert info.num_cols > 0
+    assert len(info.columns) > 0
 
     info.row_count += 1
     for i, curcol in enumerate(info.columns):
         logger.debug("tds_process_row(): reading column %d" % i)
-        curcol.funcs.get_data(tds, curcol)
+        curcol.value = curcol.funcs.get_data(tds, curcol)
     return TDS_SUCCESS
 
 # NBC=null bitmap compression row
@@ -162,7 +169,7 @@ def tds_process_nbcrow(tds):
     info = tds.current_results
     if not info:
         raise Exception('TDS_FAIL')
-    assert info.num_cols > 0
+    assert len(info.columns) > 0
     info.row_count += 1
 
     # reading bitarray for nulls, 1 represent null values for
@@ -172,7 +179,7 @@ def tds_process_nbcrow(tds):
         if ord(nbc[i/8]) & (1 << i%8):
             curcol.value = None
         else:
-            curcol.funcs.get_data(tds, curcol)
+            curcol.value = curcol.funcs.get_data(tds, curcol)
     return TDS_SUCCESS
 #
 # tds_process_end() processes any of the DONE, DONEPROC, or DONEINPROC
@@ -219,20 +226,20 @@ def tds_process_env_chg(tds):
     if type == TDS_ENV_SQLCOLLATION:
         size = r.get_byte()
         logger.debug("tds_process_env_chg(): {0} bytes of collation data received".format(size))
-        logger.debug("tds.collation was {0}".format(tds.collation))
-        tds.collation = r.get_collation()
+        logger.debug("tds.collation was {0}".format(tds.conn.collation))
+        tds.conn.collation = r.get_collation()
         r.skip(size - 5)
-        #tds7_srv_charset_changed(tds, tds.collation)
-        logger.debug("tds.collation now {0}".format(tds.collation))
+        #tds7_srv_charset_changed(tds, tds.conn.collation)
+        logger.debug("tds.collation now {0}".format(tds.conn.collation))
         # discard old one
         r.skip(r.get_byte())
     elif type == TDS_ENV_BEGINTRANS:
         size = r.get_byte()
         # TODO: parse transaction
-        tds.tds72_transaction = r.readall(8)
+        tds.conn.tds72_transaction = r.readall(8)
         r.skip(r.get_byte())
     elif type == TDS_ENV_COMMITTRANS or type == TDS_ENV_ROLLBACKTRANS:
-        tds.tds72_transaction = None
+        tds.conn.tds72_transaction = None
         r.skip(r.get_byte())
         r.skip(r.get_byte())
     elif type == TDS_ENV_PACKSIZE:
@@ -250,16 +257,16 @@ def tds_process_env_chg(tds):
     elif type == TDS_ENV_DATABASE:
         newval = r.read_ucs2(r.get_byte())
         oldval = r.read_ucs2(r.get_byte())
-        tds.env.database = newval
+        tds.conn.env.database = newval
     elif type == TDS_ENV_LANG:
         newval = r.read_ucs2(r.get_byte())
         oldval = r.read_ucs2(r.get_byte())
-        tds.env.language = newval
+        tds.conn.env.language = newval
     elif type == TDS_ENV_CHARSET:
         newval = r.read_ucs2(r.get_byte())
         oldval = r.read_ucs2(r.get_byte())
         logger.debug("server indicated charset change to \"{0}\"\n".format(newval))
-        tds.env.charset = newval
+        tds.conn.env.charset = newval
         tds_srv_charset_changed(tds, newval)
     elif type == TDS_ENV_DB_MIRRORING_PARTNER:
         newval = r.read_ucs2(r.get_byte())
@@ -361,8 +368,8 @@ def tds_process_login_tokens(tds):
             ack = r.get_byte()
             version = r.get_uint_be()
             ver['reported'] = version
-            tds.tds_version = _SERVER_TO_CLIENT_MAPPING[version]
-            if tds.tds_version == TDS71rev1:
+            tds.conn.tds_version = _SERVER_TO_CLIENT_MAPPING[version]
+            if tds.conn.tds_version == TDS71rev1:
                 tds.tds71rev1 = True
             if ver['reported'] == TDS70:
                 ver['name'] = '7.0'
@@ -388,7 +395,7 @@ def tds_process_login_tokens(tds):
             size -= 10
             if IS_TDS7_PLUS(tds):
                 product_version = 0x80000000
-                tds.product_name = r.read_ucs2(size/2)
+                tds.conn.product_name = r.read_ucs2(size/2)
             elif IS_TDS5_PLUS(tds):
                 raise NotImplementedError()
                 #tds.product_name = tds_get_string(tds, size)
@@ -398,14 +405,15 @@ def tds_process_login_tokens(tds):
             product_version = r.get_uint_be()
             # MSSQL 6.5 and 7.0 seem to return strange values for this
             # using TDS 4.2, something like 5F 06 32 FF for 6.50
-            tds_conn(tds).product_version = product_version
+            tds.conn.product_version = product_version
             logger.debug('Product version {0:x}'.format(product_version))
             # TDS 5.0 reports 5 on success 6 on failure
             # TDS 4.2 reports 1 on success and is not present of failure
             if ack == 5 or ack == 1:
                 succeed = True
-            if tds.authentication:
-                tds.authentication = None
+            if tds.conn.authentication:
+                tds.conn.authentication.close()
+                tds.conn.authentication = None
         else:
             tds_process_default_tokens(tds, marker)
         if marker == TDS_DONE_TOKEN:
@@ -732,11 +740,16 @@ def get_api_coltype(coltype):
         return DECIMAL
     elif coltype in (SYBDATETIME, SYBDATETIME4, SYBDATETIMN):
         return DATETIME
-    elif coltype in (SYBVARCHAR, SYBCHAR, SYBTEXT):
+    elif coltype in (SYBVARCHAR, SYBCHAR, SYBTEXT,
+        XSYBNVARCHAR, XSYBNCHAR, SYBNTEXT,
+        XSYBVARCHAR, XSYBCHAR):
         return STRING
     else:
         return BINARY
 
+from tds import _Results
+
+
 #/**
 # * tds7_process_result() is the TDS 7.0 result set processing routine.  It 
 # * is responsible for populating the tds->res_info structure.
@@ -756,11 +769,14 @@ def tds7_process_result(tds):
         logger.debug("no meta data")
         return TDS_SUCCESS
 
-    tds_free_all_results(tds)
+    tds.res_info = None
+    tds.param_info = None
+    tds.has_status = False
+    tds.ret_status = False
+    tds.current_results = None
     tds.rows_affected = TDS_NO_COUNT
 
-    info = tds_alloc_results(num_cols)
-    tds.current_results = info
+    tds.current_results = info = _Results()
     if tds.cur_cursor:
         tds.cur_cursor.res_info = info
         logger.debug("set current_results to cursor->res_info")
@@ -775,7 +791,8 @@ def tds7_process_result(tds):
     logger.debug("setting up {0} columns".format(num_cols))
     header_tuple = []
     for col in range(num_cols):
-        curcol = info.columns[col]
+        curcol = _Column()
+        info.columns.append(curcol)
         tds7_get_data_info(tds, curcol)
         coltype = get_api_coltype(curcol.column_type)
         precision = curcol.column_prec if hasattr(curcol, 'column_prec') else None
@@ -784,10 +801,7 @@ def tds7_process_result(tds):
     info.native_descr = tuple((col.column_name, col.column_type)
             for col in tds.res_info.columns)
     info.description = tuple(header_tuple)
-
-    # all done now allocate a row for tds_process_row to use
-    result = tds_alloc_row(info)
-    return result
+    return info 
 
 def tds_get_type_info(tds, curcol):
     r = tds._reader
@@ -816,7 +830,7 @@ def tds7_get_data_info(tds, curcol):
     tds_get_type_info(tds, curcol)
 
     # Adjust column size according to client's encoding
-    #curcol.on_server.column_size = curcol.column_size
+    #curcol.column_size = curcol.column_size
 
     # NOTE adjustements must be done after curcol.char_codec initialization
     adjust_character_column_size(tds, curcol)
@@ -827,38 +841,28 @@ def tds7_get_data_info(tds, curcol):
     #
     curcol.column_name = r.read_ucs2(r.get_byte())
 
-    logger.debug("tds7_get_data_info: \n"
-                "\tcolname = %s (%d bytes)\n"
-                "\ttype = %d (%s)\n"
-                "\tserver's type = %d (%s)\n"
-                "\tcolumn_varint_size = %d" % (
-                curcol.column_name, len(curcol.column_name), 
-                curcol.column_type, tds_prtype(curcol.column_type),
-                curcol.on_server.column_type, tds_prtype(curcol.on_server.column_type),
-                curcol.column_varint_size))
-
     return TDS_SUCCESS
 
 #
 # Adjust column size according to client's encoding 
 #
 def adjust_character_column_size(tds, curcol):
-    if is_unicode_type(curcol.on_server.column_type):
+    if is_unicode_type(curcol.column_type):
         curcol.char_codec = ucs2_codec
 
     # Sybase UNI(VAR)CHAR fields are transmitted via SYBLONGBINARY and in UTF-16
-    if curcol.on_server.column_type == SYBLONGBINARY and \
+    if curcol.column_type == SYBLONGBINARY and \
             curcol.column_usertype in (USER_UNICHAR_TYPE, USER_UNIVARCHAR_TYPE):
 
         curcol.char_codec = ucs2_codec
     # FIXME: and sybase ??
-    if not curcol.char_codec and IS_TDS7_PLUS(tds) and is_ascii_type(curcol.on_server.column_type):
-        curcol.char_codec = tds.collation.get_codec()
+    if not curcol.char_codec and IS_TDS7_PLUS(tds) and is_ascii_type(curcol.column_type):
+        curcol.char_codec = tds.conn.collation.get_codec()
 
     if not curcol.char_codec:
         return
 
-    curcol.on_server.column_size = curcol.column_size;
+    curcol.column_size = curcol.column_size;
     curcol.column_size = determine_adjusted_size(curcol.char_codec, curcol.column_size)
 
 def determine_adjusted_size(char_codec, column_size):
@@ -925,7 +929,7 @@ def tds_process_param_result_tokens(tds):
             param = _Column()
             param.column_name = name
             tds_get_type_info(tds, param)
-            param.funcs.get_data(tds, param)
+            param.value = param.funcs.get_data(tds, param)
             tds.output_params[ordinal] = param
         else:
             r.unget_byte()
